import unittest
from ziplineio import settings
from ziplineio.request import Request
from ziplineio.app import App
from ziplineio.middleware import middleware
from ziplineio.response import format_response
from ziplineio.utils import call_handler


class TestMiddleware(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        # Initialize the app
        self.app = App()

    async def test_middleware_injection(self):
        # Mock request data
        req = Request(method="GET", path="/with-middleware")

        async def my_middleware(req, **kwargs):
            return req, {"example_ctx": "some_value"}

        # Define a handler with middleware
        @self.app.get("/with-middleware")
        @middleware([my_middleware])
        async def test_handler_with_middleware(req: Request, ctx: dict):
            return {**ctx, "message": "Middleware test"}

        # Call the route
        handler, params = self.app._router.get_handler("GET", "/with-middleware")
        response = await handler(req)

        # Assertions
        self.assertEqual(response["message"], "Middleware test")
        self.assertEqual(response["example_ctx"], "some_value")

    async def test_multiple_middlewares(self):
        # Mock request data
        req = Request(method="GET", path="/with-middleware")

        async def middleware1(req, **kwargs):
            return req, {"example_ctx1": "some_value1"}

        async def middleware2(req, **kwargs):
            return req, {"example_ctx2": "some_value2"}

        async def middleware3(req, **kwargs):
            return req, {"example_ctx3": "some_value3"}

        # Define a handler with middleware
        @self.app.get("/with-middleware")
        @middleware([middleware1, middleware2, middleware3])
        async def test_handler_with_middleware(req: Request, ctx: dict):
            return {**ctx, "message": "Middleware test"}

        # Call the route
        handler, params = self.app._router.get_handler("GET", "/with-middleware")
        response = await handler(req)

        # Assertions
        self.assertEqual(response["message"], "Middleware test")
        self.assertEqual(response["example_ctx2"], "some_value2")
        self.assertEqual(len(response.keys()), 4)

    async def test_middlewares_that_throw(self):
        # Mock request data
        req = Request(method="GET", path="/with-middleware")

        async def middleware1(req, **kwargs):
            return req, {"example_ctx1": "some_value1"}

        async def middleware2(req, **kwargs):
            raise Exception("Middleware 2 failed")

        async def middleware3(req, **kwargs):
            return req, {"example_ctx3": "some_value3"}

        # Define a handler with middleware
        @self.app.get("/with-middleware")
        @middleware([middleware1, middleware2, middleware3])
        async def test_handler_with_middleware(req: Request, ctx: dict):
            return {**ctx, "message": "Middleware test"}

        # Call the route
        handler, params = self.app._router.get_handler("GET", "/with-middleware")
        response = await call_handler(handler, req=req)
        response = format_response(response, settings.DEFAULT_HEADERS)

        # Assertions
        self.assertEqual(response["status"], 500)
        self.assertEqual(
            response["body"], b"Internal server error: Middleware 2 failed"
        )

    async def test_middlewares_that_return(self):
        # Mock request data
        req = Request(method="GET", path="/with-middleware")

        async def middleware1(req, **kwargs):
            return req, {"example_ctx1": "some_value1"}

        async def middleware2(req, **kwargs):
            return {
                "message": "Hi from middleware 2",
            }

        async def middleware3(req, **kwargs):
            return req, {"example_ctx3": "some_value3"}

        # Define a handler with middleware
        @self.app.get("/with-middleware")
        @middleware([middleware1, middleware2, middleware3])
        async def test_handler_with_middleware(req: Request, ctx: dict):
            return {**ctx, "message": "Middleware test"}

        # Call the route
        handler, params = self.app._router.get_handler("GET", "/with-middleware")
        response = await call_handler(handler, req=req)

        # Assertions
        self.assertEqual(response["message"], "Hi from middleware 2")

    async def test_app_level_middleware(self):
        # Mock request data
        req = Request(method="GET", path="/with-middleware")

        async def app_level_middleware(req, **kwargs):
            return req, {"app_ctx": "some_value"}

        self.app.middleware([app_level_middleware])

        # Define a handler without explicit middleware
        @self.app.get("/with-middleware-1")
        async def test_handler_with_middleware(req: Request, ctx: dict):
            return {**ctx, "message": "Middleware test 1"}

        # Define another handler without explicit middleware
        @self.app.get("/without-middleware-2")
        async def test_handler_without_middleware(req: Request, ctx: dict):
            return {**ctx, "message": "Middleware test 2"}

        # Call the route
        handler, params = self.app._router.get_handler("GET", "/with-middleware-1")
        response = await handler(req)

        # Assertions
        self.assertEqual(response["message"], "Middleware test 1")
        self.assertEqual(response["app_ctx"], "some_value")

        # Call the route
        handler, params = self.app._router.get_handler("GET", "/without-middleware-2")
        response = await handler(req)

        # Assertions

        self.assertEqual(response["message"], "Middleware test 2")
        self.assertEqual(response["app_ctx"], "some_value")


if __name__ == "__main__":
    unittest.main()



import unittest
from ziplineio import utils


class TestUtils(unittest.TestCase):
    def test_clean_url(self):
        url = "/user/:id"
        parsed = utils.clean_url(url)

        # Assertions
        self.assertEqual(parsed, "/user/*")

    def test_match_url_pattern(self):
        pattern = "/user/:id"
        url = "/user/123"
        parsed = utils.match_url_pattern(pattern, url)

        # Assertions
        self.assertEqual(parsed, {"id": "123"})

    def test_parse_scope(self):
        scope = {
            "query_string": b"",
            "path": "/user/12",
            "headers": [(b"host", b"localhost")],
            "method": "GET",
        }
        parsed = utils.parse_scope(scope)

        # Assertions
        self.assertEqual(parsed.method, "GET")
        self.assertEqual(parsed.path, "/user/12")
        self.assertEqual(parsed.query_params, {})
        self.assertEqual(parsed.path_params, {})
        self.assertEqual(parsed.headers, {"host": "localhost"})



import re

import uuid
from typing import Any, Callable, Dict, List, Union
from ziplineio.dependency_injector import DependencyInjector, inject, injector
from ziplineio.handler import Handler
from ziplineio.middleware import middleware


class Router:
    _id: str
    _handlers: Dict[str, Dict[str, Callable]]
    _router_level_middelwares: List[Handler]
    _not_found_handler: Handler
    _sub_routers: Dict[str, "Router"]
    _prefix: str
    _injector: DependencyInjector

    def __init__(self, prefix: str = "") -> None:
        self._id = str(uuid.uuid4())
        self._handlers = {"GET": {}, "POST": {}, "PUT": {}, "DELETE": {}}
        self._router_level_middelwares = []
        self._not_found_handler = None
        self._sub_routers = {}
        self._prefix = prefix.rstrip("/")
        self._injector = injector

    def middleware(self, middlewares: List[Handler]) -> None:
        self._router_level_middelwares.extend(middlewares)

    def inject(self, service_class: Any, name: str = None) -> None:
        self._injector.add_injected_service(service_class, name, self._id)

    def _convert_path_to_regex(self, path: str) -> str:
        # Convert a path like '/user/:id' to a regex pattern like '/user/(?P<id>[^/]+)'
        return re.sub(r":(\w+)", r"(?P<\1>[^/]+)", path) + "$"

    def route(self, method: str, path: str) -> Callable[[Callable], Callable]:
        def decorator(handler: Callable) -> Callable:
            # inject router-level dependencies into the handler
            services = injector.get_injected_services(self._id)

            for name, service in services.items():
                handler = inject(service, name)(handler)

            # wrap the handler with router-level middlewares
            if len(self._router_level_middelwares) > 0:
                handler = middleware(self._router_level_middelwares)(handler)

            # Convert the path into a regex pattern
            path_regex = self._convert_path_to_regex(self._prefix + path)
            self._handlers[method][path_regex] = handler
            return handler

        return decorator

    def get(self, path: str) -> Callable[[Callable], Callable]:
        return self.route("GET", path)

    def post(self, path: str) -> Callable[[Callable], Callable]:
        return self.route("POST", path)

    def put(self, path: str) -> Callable[[Callable], Callable]:
        return self.route("PUT", path)

    def delete(self, path: str) -> Callable[[Callable], Callable]:
        return self.route("DELETE", path)

    def not_found(self, handler: Handler) -> None:
        self._not_found_handler = handler

    def add_sub_router(self, prefix: str, sub_router: "Router") -> None:
        self._sub_routers[prefix.rstrip("/")] = sub_router

    def _match_route(
        self, method: str, path: str
    ) -> Union[Callable, None, Dict[str, Any]]:
        # First, try to match in the current router
        for route_regex, handler in self._handlers[method].items():
            match = re.match(route_regex, path)
            if match:
                # Extract the path parameters
                return handler, match.groupdict()

        # Next, try to match in sub-routers
        for prefix, sub_router in self._sub_routers.items():
            if path.startswith(prefix):
                sub_path = path[len(prefix) :]
                return sub_router._match_route(method, sub_path)

        return None, {}

    def get_handler(
        self, method: str, path: str
    ) -> Union[Callable, None, Dict[str, Any]]:
        handler, params = self._match_route(method, path)
        return handler, params



from ziplineio.request import Request


def validate(*validators):
    def decorator(handler):
        async def wrapper(req: Request, ctx: dict):
            errors = {}
            validated_body = {}

            for validator in validators:
                validation_result = validator(req)
                validated_body.update(validation_result.get("validated", {}))
                errors.update(validation_result.get("errors", {}))

            if errors:
                return {"errors": errors}, 400

            req.body = validated_body
            return await handler(req, ctx)

        return wrapper

    return decorator


# # Example Pydantic model
# class UserModel(BaseModel):
#     username: str
#     age: int


# # Usage example with Pydantic model
# @app.post("/")
# @validate_body(BodyParam("user", UserModel))
# async def create_user_handler(req: Request, ctx: dict):
#     user = req.validated_body.get("user")
#     return {"user": user.dict()}  # Access the validated Pydantic model instance


# # Usage example with simple type validation
# @app.get("/")
# @validate_query(QueryParam("bar", str))
# async def test_handler(req: Request, ctx: dict):
#     bar = req.validated_query.get("bar")
#     return {"bar": bar}



import asyncio
from concurrent.futures import ProcessPoolExecutor

import functools
import inspect


# Define a top-level function to handle the sync function execution
def execute_sync(func, *args, **kwargs):
    return func(*args, **kwargs)


# Ensure the process pool executor is a singleton
class SyncExecutor:
    def __init__(self, max_workers=5):
        self.executor = ProcessPoolExecutor(max_workers=max_workers)

    async def run_in_executor(self, func, *args, **kwargs):
        loop = asyncio.get_event_loop()
        return await loop.run_in_executor(
            self.executor, functools.partial(execute_sync, func, *args, **kwargs)
        )

    def shutdown(self):
        self.executor.shutdown(wait=True)

    @classmethod
    def get_instance(cls, max_workers=5):
        if not hasattr(cls, "_instance"):
            cls._instance = cls(max_workers=max_workers)
        return cls._instance


# Factory function to create decorators for running functions in the executor
def run_sync_in_executor():
    executor = SyncExecutor()

    def decorator(func):
        async def wrapper(*args, **kwargs):
            if not inspect.iscoroutinefunction(func):
                return await executor.run_in_executor(func, *args, **kwargs)
            else:
                return await func(*args, **kwargs)

        return wrapper

    return decorator


# # def handle_exit(signum, frame):
# #     executor_instance = SyncExecutor.get_instance()
# #     executor_instance.shutdown()
# #     print("Executor shutdown")
# #     exit(0)


# # signal.signal(signal.SIGINT, handle_exit)
# # signal.signal(signal.SIGTERM, handle_exit)



from functools import wraps
import inspect
from typing import Type, Union
from pydantic import BaseModel, ValidationError
from ziplineio.request import Request


class QueryParam:
    def __init__(
        self, param: str, type: Union[Type, BaseModel] = str, required: bool = True
    ):
        self.param = param
        self.type = type
        self.required = required

    def validate(self, value):
        if isinstance(self.type, type) and issubclass(self.type, BaseModel):
            try:
                # Use Pydantic model for validation
                model_instance = self.type.parse_obj(value)
                return model_instance
            except ValidationError as e:
                raise ValueError(f"Validation error for {self.param}: {e}")

        try:
            # Basic type casting
            return self.type(value)
        except ValueError:
            raise ValueError(
                f"Invalid type for {self.param}, expected {self.type.__name__}"
            )


def validate_query(*query_params):
    """
    Decorator to validate query parameters.
    """

    def decorator(handler):
        # Use inspect to get the function signature and find the request parameter name
        signature = inspect.signature(handler)
        request_param_name = None

        # Find the parameter annotated with the Request type
        for param_name, param in signature.parameters.items():
            if param.annotation == Request or isinstance(
                param_name, Request
            ):  # Look for the parameter with type `Request`
                request_param_name = param_name
                break

        if request_param_name is None:
            raise ValueError(
                "Handler function must have a parameter of type `Request`."
            )

        @wraps(handler)
        async def wrapper(*args, **kwargs):
            req = kwargs.get(request_param_name)

            errors = {}
            validated_query = {}

            for param in query_params:
                param_name = param.param
                value = req.query_params.get(param_name)
                if value is None:
                    if param.required:
                        errors[param_name] = "Missing required query parameter"
                else:
                    try:
                        validated_query[param_name] = param.validate(value)
                    except (ValueError, ValidationError) as e:
                        errors[param_name] = str(e)

            if errors:
                return {"errors": errors}, 400

            # Attach validated query parameters to the request
            req.query_params = validated_query

            return await handler(*args, **kwargs)

        return wrapper

    return decorator if query_params else lambda req: decorator(lambda *_: None)(req)



import inspect
from typing import Any, Callable, List, Tuple, Type

from ziplineio.exception import NotFoundHttpException
from ziplineio.middleware import run_middleware_stack
from ziplineio.dependency_injector import inject, injector, DependencyInjector
from ziplineio import settings
from ziplineio.handler import Handler
from ziplineio.request import Request
from ziplineio.request_context import set_request
from ziplineio.response import Response, NotFoundResponse, format_response
from ziplineio.router import Router
from ziplineio.static import staticfiles
from ziplineio.utils import call_handler, parse_scope


class App:
    _router: Router
    _injector: DependencyInjector

    def __init__(self) -> None:
        self._router = Router()
        self._injector = injector

    def router(self, prefix: str, router: Router) -> None:
        self._router.add_sub_router(prefix, router)

    def get(self, path: str) -> Callable[[Handler], Callable]:
        return self._router.get(path)

    def post(self, path: str) -> Callable[[Handler], Callable]:
        return self._router.post(path)

    def put(self, path: str) -> Callable[[Handler], Callable]:
        return self._router.put(path)

    def delete(self, path: str) -> Callable[[Handler], Callable]:
        return self._router.delete(path)

    def not_found(self, handler: Handler) -> None:
        self._router.not_found(handler)

    def get_handler(self, method: str, path: str) -> Tuple[Handler, dict]:
        app_level_deps = self._injector.get_injected_services("app")
        handler, params = self._router.get_handler(method, path)

        if handler is None:
            return None, {}

        # TODO: Inject deps when added to the router, not here
        # inject app-level dependencies into the handler if the
        # handler expects them in its signature
        sig = inspect.signature(handler)

        filtered_kwargs_names = [
            name
            for name, param in sig.parameters.items()
            if param.default == inspect.Parameter.empty
        ]

        for name, service in app_level_deps.items():
            if "kwargs" in sig.parameters.keys() or name in filtered_kwargs_names:
                handler = inject(service, name)(handler)

        return handler, params

    def inject(
        self, service_class: Type, name: str = None
    ) -> Callable[[Callable], Callable]:
        if isinstance(service_class, list):
            for service in service_class:
                self._injector.add_injected_service(service, name, "app")
            return None
        return self._injector.add_injected_service(service_class, name, "app")

    def middleware(self, middlewares: List[Callable]) -> None:
        self._router.middleware(middlewares)

    def static(self, path: str, path_prefix: str = "/static") -> None:
        self.middleware([staticfiles(path, path_prefix)])

    async def _get_and_call_handler(
        self, method: str, path: str, req: Request
    ) -> Callable:
        # Retrieve the handler and path parameters for the given method and path
        handler, path_params = self.get_handler(method, path)
        req.path_params = path_params

        # set request context
        set_request(req)

        if handler is not None:
            # If a handler is found, call it with the request
            return await call_handler(handler, req=req)

        # If no handler was found, attempt to run middlewares.
        # (If a handler was found, middlewares will be run by `call_handler`)
        req, ctx, res = await run_middleware_stack(
            self._router._router_level_middelwares, req=req
        )

        # If middleware does not provide a response, return a 404 Not Found
        # return res if res is not None else NotFoundHttpException()
        if res is not None:
            return res

        if self._router._not_found_handler:
            response = await call_handler(self._router._not_found_handler, req=req)
            headers = isinstance(response, Response) and response._headers or {}
            return NotFoundResponse(response, headers)

        return NotFoundHttpException()

    def __call__(self, *args: Any, **kwds: Any) -> Any:
        async def uvicorn_handler(scope: dict, receive: Any, send: Any) -> None:
            if scope["type"] == "http":
                req = parse_scope(scope)
                response = await self._get_and_call_handler(req.method, req.path, req)
                raw_response = format_response(response, settings.DEFAULT_HEADERS)

                await send(
                    {
                        "type": "http.response.start",
                        "status": raw_response["status"],
                        "headers": raw_response["headers"],
                    }
                )

                await send(
                    {
                        "type": "http.response.body",
                        "body": raw_response["body"],
                    }
                )

        return uvicorn_handler



from typing import Dict


class Request:
    def __init__(
        self,
        method: str,
        path: str,
        query_params: Dict[str, str] = {},
        path_params: Dict[str, str] = {},
        headers: Dict[str, str] = {},
        body: str = "",
    ):
        self.method = method
        self.path = path
        self.path_params = path_params
        self.query_params = query_params
        self.headers = headers
        self.body = body

    method: str
    path: str
    query_params: Dict[str, str]
    path_params: Dict[str, str]
    headers: Dict[str, str]
    body: str



import json
from typing import List, Tuple, TypedDict, Dict


from ziplineio.exception import BaseHttpException


class RawResponse(TypedDict):
    headers: List[Tuple[bytes, bytes]]
    status: int
    body: bytes


class Response:
    def __init__(self, status: int, headers: Dict[str, str], body: str):
        self.status = status
        self._headers = headers
        self.body = body

    status: int
    _headers: Dict[str, str]
    body: str

    # If content-type is not provided, infer content-type from the body
    def get_headers(self) -> Dict[str, str]:
        headers = self._headers.copy()
        if "Content-Type" not in headers:
            if isinstance(self.body, bytes):
                headers["Content-Type"] = "text/plain"
            elif isinstance(self.body, str):
                headers["Content-Type"] = "text/html"
            elif isinstance(self.body, dict):
                headers["Content-Type"] = "application/json"
            else:
                headers["Content-Type"] = "text/plain"
        return headers

    def __len__(self) -> int:
        return 1


class StaticFileResponse(Response):
    def __init__(self, file_path: str, headers: Dict[str, str]):
        body = self.get_file(file_path)
        super().__init__(200, headers, body)

    def get_file(self, file_path: str) -> bytes:
        with open(file_path, "rb") as file:
            return file.read()


class JinjaResponse(Response):
    def __init__(self, body: str):
        super().__init__(200, {"Content-Type": "text/html"}, body)


class NotFoundResponse(Response):
    def __init__(self, body: str, headers: Dict[str, str] = {}):
        super().__init__(404, headers, body)


def format_headers(headers: Dict[str, str] | None) -> List[Tuple[bytes, bytes]]:
    if headers is None:
        return []
    return [(bytes(k, "utf-8"), bytes(v, "utf-8")) for k, v in headers.items()]


def format_body(body: bytes | str | dict) -> bytes:
    if isinstance(body, bytes):
        return body
    elif isinstance(body, str):
        return bytes(body, "utf-8")
    elif isinstance(body, dict):
        return bytes(json.dumps(body), "utf-8")
    elif isinstance(body, Response):
        return format_body(body.body)
    raise ValueError("Invalid body type")


def format_response(
    response: bytes | dict | str | Response | Exception, default_headers: dict[str, str]
) -> RawResponse:
    # Helper to merge and deduplicate headers

    # Format different response types
    if isinstance(response, bytes):
        headers = [(b"content-type", b"text/plain")]
        body = response
        status = 200

    elif isinstance(response, str):
        headers = [(b"content-type", b"text/plain")]
        body = response.encode("utf-8")
        status = 200

    elif isinstance(response, dict):
        headers = [(b"content-type", b"application/json")]
        body = json.dumps(response).encode("utf-8")
        status = 200

    elif isinstance(response, Response):
        headers = format_headers(response.get_headers())
        body = format_body(response.body)
        status = response.status

    elif isinstance(response, BaseHttpException):
        headers = [(b"content-type", b"application/json")]
        body = format_body(response.message)
        status = response.status_code

    elif isinstance(response, Exception):
        headers = [(b"content-type", b"text/plain")]
        body = f"Internal server error: {str(response)}".encode("utf-8")
        status = 500

    else:
        raise ValueError("Invalid response type")

    default_headers = format_headers(default_headers)

    return {"headers": default_headers + headers, "status": status, "body": body}



import inspect
from typing import Any
from ziplineio.response import JinjaResponse
from ziplineio.utils import call_handler


def jinja(env: Any, template_name: str):
    template = env.get_template(template_name)

    def decorator(handler):
        async def wrapped_handler(req, **kwargs):
            # Pass all arguments directly to the handler
            # sig = inspect.signature(handler)

            # Filter kwargs to only pass those that the handler expects
            # filtered_kwargs = {k: v for k, v in kwargs.items() if k in sig.parameters}
            context = await call_handler(handler, req=req, **kwargs)
            rendered = template.render(context)
            return JinjaResponse(rendered)

        return wrapped_handler

    return decorator



from concurrent.futures import ProcessPoolExecutor
import unittest
import asyncio
from unittest.mock import patch

from ziplineio.process_pool import SyncExecutor, run_sync_in_executor


# Standalone synchronous functions that can be pickled
def sync_function(x, y):
    return x + y


def another_sync_function(x):
    return x * 2


class TestSyncExecutor(unittest.TestCase):
    def test_sync_function_in_executor(self):
        wrapped_sync_function = run_sync_in_executor()(sync_function)
        result = asyncio.run(wrapped_sync_function(5, 10))
        self.assertEqual(result, 15)

    def test_another_sync_function_in_executor(self):
        wrapped_another_sync_function = run_sync_in_executor()(another_sync_function)
        result = asyncio.run(wrapped_another_sync_function(3))
        self.assertEqual(result, 6)

    @patch.object(SyncExecutor, "get_instance", wraps=SyncExecutor.get_instance)
    def test_executor_is_shared_across_calls(self, mock_get_instance):
        wrapped_sync_function = run_sync_in_executor()(sync_function)
        wrapped_another_sync_function = run_sync_in_executor()(another_sync_function)

        result1 = asyncio.run(wrapped_sync_function(5, 10))
        result2 = asyncio.run(wrapped_another_sync_function(3))

        self.assertEqual(result1, 15)
        self.assertEqual(result2, 6)

        # Check if the same executor instance is returned both times
        instance1 = mock_get_instance()
        instance2 = mock_get_instance()

        self.assertIs(
            instance1,
            instance2,
            "The same executor instance should be used across calls",
        )

    @patch.object(ProcessPoolExecutor, "shutdown")
    def test_executor_shutdown(self, mock_shutdown):
        wrapped_sync_function = run_sync_in_executor()(sync_function)
        asyncio.run(wrapped_sync_function(2, 3))

        # Manually trigger shutdown and verify
        executor_instance = SyncExecutor.get_instance()
        executor_instance.shutdown()

        mock_shutdown.assert_called_once()


if __name__ == "__main__":
    unittest.main()



import unittest

from pydantic import BaseModel


from ziplineio.app import App

from ziplineio.request import Request

from ziplineio.utils import call_handler
from ziplineio.validation.body import BodyParam, validate_body
from ziplineio.validation.query import QueryParam, validate_query


class TestValidateBody(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        # Initialize the app
        self.app = App()

    async def test_validate_simple_body_params(self):
        @self.app.post("/")
        @validate_body(BodyParam("username", str), BodyParam("age", float))
        async def create_user_handler(req: Request):
            username = req.body.get("username")
            age = req.body.get("age")
            return {"username": username, "age": age}

        req = Request(method="POST", path="/", body={"username": "Amy", "age": 0.11})
        response = await call_handler(create_user_handler, req=req)

        self.assertEqual(response["username"], "Amy")
        self.assertEqual(response["age"], 0.11)

    async def test_validate_pydantic_body_params(self):
        # # Example Pydantic model
        class UserModel(BaseModel):
            username: str
            age: int

        @self.app.post("/")
        @validate_body(BodyParam("user", UserModel))
        async def create_user_handler(req: Request):
            user = req.body.get("user")
            return {
                "user": user.model_dump()
            }  # Access the validated Pydantic model instance

        req = Request(
            method="POST", path="/", body={"user": {"username": "John", "age": 30}}
        )
        response = await call_handler(create_user_handler, req=req)

        self.assertEqual(response["user"], {"username": "John", "age": 30})

    async def test_validate_missing_required_body_params(self):
        @self.app.post("/")
        @validate_body(BodyParam("username", str), BodyParam("age", float))
        async def create_user_handler(req: Request):
            username = req.body.get("username")
            age = req.body.get("age")
            return {"username": username, "age": age}

        req = Request(method="POST", path="/", body={"username": "Amy"})
        response = await call_handler(create_user_handler, req=req)

        self.assertEqual(
            response, ({"errors": {"age": "Missing required body parameter"}}, 400)
        )

    async def test_validate_missing_optional_body_params(self):
        @self.app.post("/")
        @validate_body(
            BodyParam("username", str, required=False), BodyParam("age", float)
        )
        async def create_user_handler(req: Request):
            username = req.body.get("username")
            age = req.body.get("age")
            return {"username": username, "age": age}

        req = Request(method="POST", path="/", body={"age": 0.11})
        response = await call_handler(create_user_handler, req=req)

        self.assertEqual(response, {"username": None, "age": 0.11})

    async def test_validate_invalid_body_params(self):
        @self.app.post("/")
        @validate_body(BodyParam("username", str), BodyParam("age", float))
        async def create_user_handler(req: Request):
            username = req.body.get("username")
            age = req.body.get("age")
            return {"username": username, "age": age}

        req = Request(
            method="POST", path="/", body={"username": "Amy", "age": "invalid"}
        )
        response = await call_handler(create_user_handler, req=req)

        self.assertEqual(
            response,
            (
                {"errors": {"age": "Invalid type for age, expected float"}},
                400,
            ),
        )


class TestValidateQuery(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        # Initialize the app
        self.app = App()

    async def test_validate_simple_query_params(self):
        @self.app.get("/")
        @validate_query(QueryParam("username", str), QueryParam("age", float))
        async def create_user_handler(req: Request):
            username = req.query_params.get("username")
            age = req.query_params.get("age")
            return {"username": username, "age": age}

        req = Request(
            method="GET", path="/", query_params={"username": "Amy", "age": 0.11}
        )
        response = await call_handler(create_user_handler, req=req)

        self.assertEqual(response["username"], "Amy")
        self.assertEqual(response["age"], 0.11)

    async def test_validate_pydantic_query_params(self):
        # # Example Pydantic model
        class UserModel(BaseModel):
            username: str
            age: int

        @self.app.get("/")
        @validate_query(QueryParam("user", UserModel))
        async def create_user_handler(req: Request):
            user = req.query_params.get("user")
            return {"user": user.model_dump()}

        req = Request(
            method="GET",
            path="/",
            query_params={"user": {"username": "John", "age": 30}},
        )
        response = await call_handler(create_user_handler, req=req)

        self.assertEqual(response["user"], {"username": "John", "age": 30})

    async def test_validate_missing_required_query_params(self):
        @self.app.get("/")
        @validate_query(QueryParam("username", str), QueryParam("age", float))
        async def create_user_handler(req: Request):
            username = req.query_params.get("username")
            age = req.query_params.get("age")
            return {"username": username, "age": age}

        req = Request(method="GET", path="/", query_params={"username": "Amy"})
        response = await call_handler(create_user_handler, req=req)

        self.assertEqual(
            response, ({"errors": {"age": "Missing required query parameter"}}, 400)
        )

    async def test_validate_missing_optional_query_params(self):
        @self.app.get("/")
        @validate_query(
            QueryParam("username", str, required=False), QueryParam("age", float)
        )
        async def create_user_handler(req: Request):
            username = req.query_params.get("username")
            age = req.query_params.get("age")
            return {"username": username, "age": age}

        req = Request(method="GET", path="/", query_params={"age": 0.11})
        response = await call_handler(create_user_handler, req=req)

        self.assertEqual(response, {"username": None, "age": 0.11})

    async def test_validate_invalid_query_params(self):
        @self.app.get("/")
        @validate_query(QueryParam("username", str), QueryParam("age", float))
        async def create_user_handler(req: Request):
            username = req.query_params.get("username")
            age = req.query_params.get("age")
            return {"username": username, "age": age}

        req = Request(
            method="GET", path="/", query_params={"username": "Amy", "age": "invalid"}
        )
        response = await call_handler(create_user_handler, req=req)

        self.assertEqual(
            response,
            (
                {"errors": {"age": "Invalid type for age, expected float"}},
                400,
            ),
        )



import unittest

from ziplineio import settings
from ziplineio.app import App
from ziplineio.exception import BaseHttpException
from ziplineio.response import format_response
from ziplineio.utils import call_handler


class TestExceptions(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        # Initialize the app
        self.app = App()

    async def test_base_exception(self):
        # Inject the service
        @self.app.get("/exception")
        async def test_handler(req):
            raise BaseHttpException("Hey! We messed up", 409)

        # Call the handler
        response = await call_handler(test_handler, req={})
        response = format_response(response, settings.DEFAULT_HEADERS)
        self.assertEqual(response["body"], b"Hey! We messed up")
        self.assertEqual(response["status"], 409)

    async def test_exception_extends_base_exception(self):
        class CustomHttpException(BaseHttpException):
            pass

        @self.app.get("/exception")
        async def test_handler(req):
            raise CustomHttpException("Hey! We messed up bad", 402)

        # Call the handler
        response = await call_handler(test_handler, req={})
        response = format_response(response, settings.DEFAULT_HEADERS)
        self.assertEqual(response["body"], b"Hey! We messed up bad")
        self.assertEqual(response["status"], 402)
